;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Decompresses some PeX 0.99 files as well as all known public versions of Continuum.
;; Written by Snrrrub (snrrrub@yahoo.com): April 16, 2004
;; Original DeX code (c) by Chafe 2000
;; 

.386
.model flat, stdcall
option casemap:none

include C:\masm32\include\windows.inc
include C:\masm32\include\kernel32.inc
include C:\masm32\include\comdlg32.inc

includelib C:\masm32\lib\kernel32.lib
includelib C:\masm32\lib\comdlg32.lib

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; IMPORT FUNCTIONS
;;

WinMain				PROTO	:HINSTANCE, :HINSTANCE, :LPSTR, :DWORD

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; EXPORT FUNCTIONS
;;

public unpack
public FreeAllStuff

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; STRUCTURE DEFINITIONS AND CONSTANTS
;;

VERSIONS				equ 5					; The number of versions we support
VersionInfo struct
	index				dd  ?					; The integer index of this item in the VersionInfo array
	versionName			db  128 dup (0)			; String identifying this version of PeX
	checksum			dd  ?					; The checksum of the first 40 dwords
	DecryptAlgo			dd  ?					; Address of decryption algorithm in executable
	EncryptData			dd  ?					; Address of encrypted data
	UnpackedSize		dd  ?					; Address of the total uncompressed size of compressed sections
	Sections			dd  ?					; Address (in second stage) of the previously encrypted sections
	EntryPoint			dd  ?					; Address (in second stage) of the entry point of the executable
	DecompAlgo			dd  ?
	ImportTable			dd  ?
	ImportKey			dd  ?
	ImportROL			dd  ?
	SizeFunction		dd  ?
	FixSectionFunction	dd	?
VersionInfo ends

.data
COMM C m_inputFile			:DWORD
COMM C m_outputFile			:DWORD
COMM C m_outputFileSize		:DWORD
COMM C m_decryptedData		:DWORD
COMM C m_decryptedSize		:DWORD
COMM C PEOffset				:DWORD
COMM C SectTable			:DWORD
COMM C PEXSection			:DWORD
COMM C Version				:DWORD

hInstance			dd		0
Versions            VersionInfo { 0, 'PeX v0.99',       03083fef2h, 1abh, 3feh, 214h, 24h, 397h, 25ch, 8dh, 19bh, 1c7h, CalculateSize, FixSection }
					VersionInfo	{ 1, 'Ctm v0.35/v0.36', 0e86b0cf7h, 1abh, 3feh, 214h, 24h, 397h, 25ch, 8dh, 19bh, 1c7h, CalculateSize, FixSection }
					VersionInfo	{ 2, 'Ctm v0.37',       0ef1a2b1bh, 0b1h, 317h, 118h, 24h, 3e4h, 160h, 96h, 1a4h, 1d0h, CTM_CalculateSize, CTM_FixSection }
					VersionInfo	{ 3, 'Ctm v0.38',       0fe0182dah, 0b1h, 317h, 118h, 24h, 3e4h, 160h, 96h, 1a4h, 1d0h, CTM_CalculateSize, CTM_FixSection }
					VersionInfo	{ 4, 'Ctm v0.39pr1',    0eb89f207h, 0b1h, 317h, 118h, 24h, 3e4h, 160h, 96h, 1a4h, 1d0h, CTM_CalculateSize, CTM_FixSection }

.CODE
start:
	invoke	GetModuleHandle, NULL
	mov		hInstance, eax

	invoke	GetCommandLine
	invoke	WinMain, hInstance, NULL, eax, SW_SHOW
	invoke	ExitProcess, eax

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Returns a pointer to the PE header within the executable.
;;
getPEOffset PROC C, mzFile:PTR BYTE, fileSize:DWORD
	push	ecx
	mov		eax, mzFile
	mov		ecx, eax
	add		ecx, fileSize
	sub		ecx, 4h
	add		eax, dword ptr [eax + 3ch]		; get offset for PE header within the MZ file
	cmp		eax, ecx
	jge		@error							; PE header is outside file
	cmp		dword ptr [eax], 4550h			; Make sure PE header exists
	je		@done
@error:
	xor		eax, eax
@done:
	pop		ecx
	ret
getPEOffset ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Returns a pointer to the section table given a pointer to a PE header.
;;
getSectionTable PROC C, peFile:PTR BYTE
	push	edx
	mov		eax, peFile
	movzx	edx, word ptr [eax + 14h]
	add		eax, edx
	add		eax, 18h
	pop		edx
	ret
getSectionTable ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Returns the RVA of the entry point of a PE file given a pointer to the start of the file.
;;
getEntryPointRVA PROC C, peFile:PTR BYTE
	mov		eax, peFile
	mov		eax, dword ptr [eax + 28h]
	ret
getEntryPointRVA ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Returns a pointer to the section table entry that contains the given (RVA) address.
;; WARNING: the address MUST exist in one of the section table entries.
;;
findSectionEntryWithRVA PROC C, sectionTable:PTR BYTE, address:DWORD
	push	ecx
	push	edx
	mov		eax, sectionTable
	mov		edx, address
@tryAgain:
	mov		ecx, dword ptr [eax + 0ch]
	cmp		ecx, edx
	jg		@notHere
	add		ecx, dword ptr [eax + 8h]
	cmp		ecx, edx
	jg		@found
@notHere:
	add		eax, 28h
	jmp		@tryAgain
@found:
	pop		edx
	pop		ecx
	ret
findSectionEntryWithRVA ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Adds the first 40 DWORDS of a PeX section as a checksum. Requires a pointer to the PeX section.
;;
getPhysicalOffset PROC C, sectionEntry:PTR BYTE
	mov		eax, sectionEntry
	mov		eax, dword ptr [eax + 14h]
	ret
getPhysicalOffset ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Adds the first 40 DWORDS of a PeX section as a checksum. Requires a pointer to the PeX section.
;;
generateChecksum PROC C, pexSection:PTR BYTE
	push	ebx
	push	ecx
	xor		eax, eax
	mov		ebx, pexSection
	mov		ecx, 40h
@GC_NextDword:
	add		eax, dword ptr [ebx]
	add		ebx, 4h
	loop	@GC_NextDword
	pop		ecx
	pop		ebx
	ret
generateChecksum ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Returns a pointer into the VersionInfo array that matches the checksum. If no version matches
;; the checksum, it returns 0.
;;
findVersion PROC C, checksum:DWORD
	push	ecx
	push	edx
	mov		ecx, VERSIONS		; ecx = # of versions supported
	mov		eax, offset Versions
@nextVersion:
	mov		edx, dword ptr [eax].VersionInfo.checksum
	cmp		edx, checksum
	je		@found
	add		eax, sizeof(VersionInfo)
	dec		ecx
	jne		@nextVersion
	xor		eax, eax
@found:
	pop		edx
	pop		ecx
	ret
findVersion ENDP

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Decrypt the compressed data, decompress it, fix the sections & the header, 
;; fix the imports, guess some sectionnames, show save-dialogbox, write output..
;; And more...
;;

unpack:
    push    edi							; Save a few registers to make the caller happy
    push    ebp
    push    esi
    push    ebx
    mov     edi, PEXSection				; Since encryption-methods may be different in different
    mov     ecx, edi					; versions, I decided to place a 0C3h-byte right after the
    mov     ebx, Version				; decryption-algo in memory and then make a call to it.
    add     ecx, dword ptr [ebx].VersionInfo.DecryptAlgo
    add     edi, dword ptr [ebx].VersionInfo.EncryptData
    mov     edx, ecx
    mov     esi, edi
PatchDecryptAlgo:
    cmp     dword ptr [ecx], 49470788h	; Last bytes in the decryption-algo
    je      FoundPlaceToPatch
    inc     ecx
    jmp     PatchDecryptAlgo
FoundPlaceToPatch:
    mov     byte ptr [ecx + 6], 0c3h	; RET-instruction.. Now we could make a CALL to it.
    call    edx							; edi -> Data to decrypt. edx -> Decryption-algo..
    mov     eax, dword ptr [ebx].VersionInfo.UnpackedSize
    add     eax, PEXSection				; eax -> Size of unpacked data 
    mov     ecx, dword ptr [eax]
    mov     m_decryptedSize, ecx
    invoke  GlobalAlloc, GMEM_FIXED or GMEM_ZEROINIT, ecx
    mov     m_decryptedData, eax
    push    eax
    push    esi
    mov     eax, dword ptr [ebx].VersionInfo.DecompAlgo
    add     eax, PEXSection			; Since the decompression algorithm may be different
    call    eax						; in other versions, we just call the algo found in 
    add     esp, 8					; the file.
    xor     ecx, ecx
    mov     edx, PEOffset			; Mission: Calculate the size of the unpacked file.
    xor     eax, eax				; Method: Add the size of the header (= offset to the
    mov     cx,  word ptr [edx + 6]	; first physical section) with the physical sizes of
    dec     eax						; all sections but the one containing the PeX-code.
    push    ecx
    mov     edx, SectTable			; Step one: Find the first physical section.
FindFirstPhysicSection:
    mov     ebp, dword ptr [edx + 14h]
    test    ebp, ebp
    je      NotFirstPhysicSection	; PhysOffset == 0? Then this is a virtual section.
    cmp     eax, ebp			
    jb      NotFirstPhysicSection	
    mov     eax, ebp
NotFirstPhysicSection:
    add     edx, 28h
    dec     ecx						; Go though all sections.
    jne     FindFirstPhysicSection
    pop     ecx						; Step two: Add the sizes of the sections with the headerSize
    dec     ecx						; Exclude PeX-section when adding the sizes..

	push	eax						; save header size for a bit... (used below)
	push	eax						; push header size
	push	ecx						; push number of sections
	call	dword ptr [ebx].VersionInfo.SizeFunction
    mov     m_outputFileSize, eax			; Mission successful!

    invoke  GlobalAlloc, GMEM_FIXED or GMEM_ZEROINIT, eax

    mov     m_outputFile, eax				; This is going to be the file we write
    pop		ecx								; grab the headerSize that we pushed above... 
    mov     esi, m_inputFile
    mov     edi, eax
    mov     eax, PEOffset
    dec     word ptr [eax+6]				; Decrease the number of sections
    rep     movsb
    mov     eax, SectTable					; eax = m_inputFile's sectiontable
    mov     ebp, eax
    sub     ebp, m_inputFile
    add     ebp, m_outputFile				; ebp = m_outputFile's sectiontable
    mov     esi, m_decryptedData
    mov     ecx, dword ptr [ebx].VersionInfo.Sections
    add     esi, dword ptr [esi + ecx]	; esi -> First decompressed section
FixNextSection:
	push	ebp
	push	eax
	call	dword ptr [ebx].VersionInfo.FixSectionFunction
    add     eax, 28h
    add     ebp, 28h
    mov     ecx,PEOffset
    dec     word ptr [ecx + 6]			; Make sure all sections are fixed..
    jne     FixNextSection
    mov     eax, dword ptr [ebp + 8]	; Get VirtSize of the PeX-section
    mov     edi, PEOffset
    sub     edi, m_inputFile
    add     edi, m_outputFile
    sub     dword ptr [edi + 50h], eax	; Correct ImageSize by subtracting it
    mov     eax, dword ptr [ebx].VersionInfo.EntryPoint
    add     eax, m_decryptedData
    mov     eax, dword ptr [eax]		; eax = Entrypoint VA - 1
    sub     eax, dword ptr [edi + 34h]	; VA - ImageBase == RVA, thus eax = Entrypoint RVA - 1 
    inc     eax							; eax = Entrypoint RVA
    mov     dword ptr [edi + 28h], eax	; Update header
    xor     eax, eax
    push    edi
    mov     edi, ebp
    mov     ecx, 10
    rep     stosd						; Remove the PeX-section's entry in the sectiontable
    mov     eax, dword ptr [ebx].VersionInfo.ImportTable
    add     eax, m_decryptedData		; Find out where the import-table begins
    pop     edi
    mov     eax, dword ptr [eax]
    mov     dword ptr [edi + 80h], eax	; Store it in the header
    call    GetOffsetFromRVA
    mov     esi, eax				; Simple decryption-algorithm for the names of all
DecryptNextLibrary:    				; imported functions.
    mov     eax, dword ptr [esi]	; Study/trace it if you really want to know how it works
    test    eax, eax
    je      FinishedImports
    call    GetOffsetFromRVA
    mov     edi, eax
DecryptNextImport:
    mov     eax, dword ptr [edi]
    not     eax
    mov     dword ptr [edi], eax
    test    eax, eax
    je      FinishedLibrary
    test    eax, 80000000h
    jne     DontTouch
    call    GetOffsetFromRVA
    mov     edx, eax
    mov     eax, dword ptr [ebx].VersionInfo.ImportKey
    mov     ecx, dword ptr [ebx].VersionInfo.ImportROL
    add     eax, m_decryptedData
    add     ecx, m_decryptedData
    mov     eax, dword ptr [eax]		; Get decryptionkey from decompressed data
    mov     cl, byte ptr [ecx]			; Get number of times to ROL eax too.
    push    edx
ImportDecrypt:
    mov     ch, byte ptr [edx]
    xor     ch, al
    add     ch, ah
    sub     ch, al
    rol     eax, cl
    mov     byte ptr [edx], ch
    inc     edx
    test    ch, ch
    jne     ImportDecrypt
    pop     ecx
    dec     edx
MoveNameOfImport:    
    mov     al, byte ptr [edx]		; Move name of import two bytes to make place for a hint
    mov     byte ptr [edx + 2], al
    dec     edx
    cmp     edx, ecx
    jge     MoveNameOfImport
    mov     word ptr [ecx], 0		; Ehm.. Hints are not needed anyway.. 
DontTouch:    
    add     edi, 4
    jmp     DecryptNextImport
FinishedLibrary:
    add     esi, 14h
    jmp     DecryptNextLibrary
FinishedImports:
    xor     ecx, ecx
    mov     edx, PEOffset
    sub     edx, m_inputFile
    add     edx, m_outputFile
    and     dword ptr [edx + 0f4h], 0fffffff7h	; Remove PeX's signature
    mov     cx, word ptr [edx + 6]				; Now we're going to guess the names of the
    mov     esi, SectTable						; sections. As help, we have some values in the
    sub     esi, m_inputFile					; header. If a name can't be guessed, it'll be
    add     esi, m_outputFile					; named "DeX#", where # is a digit.
    mov     ebp, 30h							; 30h == '0'
GuessSectionNames:
    mov     eax, dword ptr [esi + 0ch]
    cmp     eax, dword ptr [edx + 80h]
    jne     NotImportSection
    mov     dword ptr [esi], 6164692eh			; Import-section - ".idata"
    mov     dword ptr [esi + 4], 6174h
    jmp     GuessNextSection
NotImportSection:
    cmp     eax,dword ptr [edx + 88h]
    jne     NotResourceSection
    mov     dword ptr [esi], 7273722eh			; Resource-section - ".rsrc"
    mov     dword ptr [esi + 4], 63h
    jmp     GuessNextSection
NotResourceSection:
    cmp     eax, dword ptr [edx + 0c0h]
    jne     NotTLSSection
    mov     dword ptr [esi], 736c742eh			; ThreadLocalStorage - ".tls"
    jmp     GuessNextSection
NotTLSSection:
    mov     ebx,dword ptr [edx + 2ch]
    cmp     eax, ebx
    jl      NotCodeSection
    add     ebx, dword ptr [edx + 1ch]
    cmp     eax, ebx
    jge     NotCodeSection
    mov     dword ptr [esi], 7865742eh			; Code-section - ".text"
    mov     dword ptr [esi + 4], 74h
    jmp     GuessNextSection
NotCodeSection:
    mov     ebx, dword ptr [edx + 30h]
    cmp     eax, ebx
    jl      NotDataSection
    add     ebx, dword ptr [edx + 20h]
    cmp     eax, ebx
    jge     NotDataSection
    mov     dword ptr [esi], 7461642eh			; m_decryptedData-section - ".data"
    mov     dword ptr [esi + 4], 61h
    jmp     GuessNextSection
NotDataSection:
    cmp     dword ptr [esi + 10h], 0
    jne     NotBSSSection
    mov     dword ptr [esi], 7373622eh			; BSS-section - ".bss"
    jmp     GuessNextSection
NotBSSSection:
    mov     word ptr [esi + 3], bp				; No idea - "DeX#"
    inc     ebp
GuessNextSection:
	add     esi, 28h
	dec     ecx
	jne     GuessSectionNames
	pop     ebx
	pop     esi
	pop     ebp
	pop     edi
	ret

GetOffsetFromRVA:						; Find the address in the output-buffer that matches
    mov     ecx, SectTable				; the RVA specified in eax. This is done by consulting
    sub     ecx, m_inputFile			; the sectiontable..
    add     ecx, m_outputFile
FindSection:
    mov     edx, dword ptr [ecx + 0ch]
    cmp     eax, edx
    jl      WrongSection
    add     edx, dword ptr [ecx + 8]
    cmp     eax, edx
    jl      FoundSection
WrongSection:
    add     ecx, 28h
    jmp     FindSection
FoundSection:
    sub     eax, dword ptr [ecx + 0ch]
    add     eax, dword ptr [ecx + 14h]
    add     eax, m_outputFile
    ret

FreeAllStuff:				; Free all (if any) allocated buffers..
    push    esi
    mov     esi, GlobalFree
    mov     eax, m_outputFile
    test    eax, eax
    je      DontFreeOutput
    push    eax
    call    esi
    mov     m_outputFile, eax
DontFreeOutput:
    mov     eax, m_decryptedData
    test    eax, eax
    je      DontFreeData
    push    eax
    call    esi
DontFreeData:
    pop     esi
    ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ESP + 0: return address
;; ESP + 4: # of sections
;; ESP + 8: header size

CalculateSize:
	push	ecx
    push    edx
    mov     edx, SectTable
    mov		ecx, dword ptr [esp + 0Ch]		; # of sections
    mov		eax, dword ptr [esp + 10h]		; header size
@CS_NextSection:
    cmp     byte ptr [edx + 22h], 0			; Is section packed (# of lines field == 0)
    jne     @CS_NotPacked
    add     eax, dword ptr [edx]			; If so, physical size is at offset 0 in section table
    jmp     @CS_IsPacked
@CS_NotPacked:
    add     eax, dword ptr [edx + 10h]		; Otherwise it's in the "standard" location
@CS_IsPacked:
    add     edx, 28h
    dec     ecx
    jne     @CS_NextSection					; Calculate size for next section now
    pop     edx
    pop		ecx
    ret		8h

CTM_CalculateSize:
	push    esi
	push	ecx
	push    edx
	mov     esi, m_decryptedData
	mov     edx, dword ptr [ebx].VersionInfo.Sections
	add     esi, dword ptr [esi + edx]
	mov     edx, SectTable
	mov		ecx, dword ptr [esp + 10h]		; # of sections
	mov		eax, dword ptr [esp + 14h]		; header size
@CTM_CS_NextSection:
	cmp     byte ptr [edx + 22h], 0
	jne		@CTM_CS_NotPacked
	add     eax, dword ptr [esi]
	add		esi, dword ptr [esi]			; Skip over the section
	add     esi, 5							; and the size  + ' ' fields
	jmp     @CTM_CS_IsPacked
@CTM_CS_NotPacked:
	add     eax, dword ptr [edx + 10h]
@CTM_CS_IsPacked:
	add     edx, 28h
	dec     ecx
	jne     @CTM_CS_NextSection
	pop     edx
	pop		ecx
	pop     esi
	ret		08h

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ESI    : pointing to first unpacked section
;; EDI    : pointing to output buffer just after the PE header has been copied
;; ESP + 0: return address
;; ESP + 4: source section table entry
;; ESP + 8: destination section table entry

FixSection:
	push	eax
	push	ebx
	push	ecx
	push	edx
	mov		eax, dword ptr [esp + 14h]			; eax = source section table entry
	mov		edx, dword ptr [esp + 18h]			; edx = destination section table entry
	cmp		byte ptr [eax + 22h], 0				; is the section packed?
	jne		@FS_NotPacked
	mov		ecx, dword ptr [eax]				; ecx = new physical size
	mov		dword ptr [edx + 10h], ecx
	mov		dword ptr [edx + 14h], edi			; Physical offset is where we're writing to
	mov		ebx, m_outputFile
	sub		dword ptr [edx + 14h], ebx			; but we need to adjust it by the base
	mov		dword ptr [edx], 586544h			; section name = "DeX" temporarily
	mov		dword ptr [edx + 4h], 0
	or		dword ptr [edx + 24h], 20000020h	; Make sure some flags are set
	rep		movsb								; copy section from unpacked data
	jmp		@FS_DoneSection
@FS_NotPacked:
	push	esi
	mov		esi, dword ptr [eax + 14h]			; get the offset
	add		esi, m_inputFile							; from the input file
	mov		ecx, dword ptr [eax + 10h]			; and the size of the section
	mov		dword ptr [edx + 14h], edi			; rewrite the physical offset
	mov		ebx, m_outputFile
	sub		dword ptr [edx + 14h], ebx			; and fix it up
	mov		dword ptr [edx], 586544h			; section name = "DeX" temporarily
	mov		dword ptr [edx + 4h], 0
	or		dword ptr [edx + 24h], 20000020h	; Make sure some flags are set
	rep		movsb								; copy section from original input file
	pop		esi
@FS_DoneSection:
	pop		edx
	pop		ecx
	pop		ebx
	pop		eax
	ret		08h

CTM_FixSection:
	push	eax
	push	ebx
	push	ecx
	push	edx
	mov		eax, dword ptr [esp + 14h]			; eax = source section table entry
	mov		edx, dword ptr [esp + 18h]			; edx = destination section table entry
	cmp		byte ptr [eax + 22h], 0				; is the section packed?
	jne		@CTM_FS_NotPacked
	mov		ecx, dword ptr [esi]
	add		esi, 5
	mov		dword ptr [edx + 10h], ecx
	mov		dword ptr [edx + 14h], edi			; Physical offset is where we're writing to
	mov		ebx, m_outputFile
	sub		dword ptr [edx + 14h], ebx			; but we need to adjust it by the base
	mov		dword ptr [edx], 586544h			; section name = "DeX" temporarily
	mov		dword ptr [edx + 4h], 0
	or		dword ptr [edx + 24h], 20000020h	; Make sure some flags are set
	rep		movsb								; copy section from unpacked data
	jmp		@CTM_FS_DoneSection
@CTM_FS_NotPacked:
	push	esi
	mov		esi, dword ptr [eax + 14h]			; get the offset
	add		esi, m_inputFile							; from the input file
	mov		ecx, dword ptr [eax + 10h]			; and the size of the section
	mov		dword ptr [edx + 14h], edi			; rewrite the physical offset
	mov		ebx, m_outputFile
	sub		dword ptr [edx + 14h], ebx			; and fix it up
	mov		dword ptr [edx], 586544h			; section name = "DeX" temporarily
	mov		dword ptr [edx + 4h], 0
	or		dword ptr [edx + 24h], 20000020h	; Make sure some flags are set
	rep		movsb								; copy section from original input file
	pop		esi
@CTM_FS_DoneSection:
	pop		edx
	pop		ecx
	pop		ebx
	pop		eax
	ret		08h

end start